/*
 * Copyright (c) 1997 - 2004 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "gen_locl.h"

RCSID("$Id: gen_decode.c,v 1.12.2.29 2004/11/10 16:16:05 lha Exp $");

static void
decode_primitive (const char *typename, const char *name, const char *forwstr)
{
#if 0
    fprintf (codefile,
	     "e = decode_%s(p, len, %s, &l);\n"
	     "%s;\n",
	     typename,
	     name,
	     forwstr);
#else
    fprintf (codefile,
	     "e = der_get_%s(p, len, %s, &l);\n"
	     "if(e) %s;\np += l; len -= l; ret += l;\n",
	     typename,
	     name,
	     forwstr);
#endif
}

static int
is_primitive_type(int type)
{
    switch(type) {
    case TInteger:
    case TBoolean:
    case TOctetString:
    case TBitString:
    case TEnumerated:
    case TGeneralizedTime:
    case TGeneralString:
    case TOID:
    case TUTCTime:
    case TUTF8String:
    case TPrintableString:
    case TIA5String:
    case TNull:
	return 1;
    default:
	return 0;
    }
}

static int
decode_type (const char *name, const Type *t, int optional, 
	     const char *forwstr)
{
    switch (t->type) {
    case TType: {
	if (optional)
	    fprintf(codefile, 
		    "%s = calloc(1, sizeof(*%s));\n"
		    "if (%s == NULL) return ENOMEM;\n",
		    name, name, name);
	fprintf (codefile,
		 "e = decode_%s(p, len, %s, &l);\n",
		 t->symbol->gen_name, name);
	if (optional) {
	    fprintf (codefile,
		     "if(e) {\n"
		     "free(%s);\n"
		     "%s = NULL;\n"
		     "} else {\n"
		     "p += l; len -= l; ret += l;\n"
		     "}\n",
		     name, name);
	} else {
	    fprintf (codefile,
		     "if(e) %s;\n",
		     forwstr);
	    fprintf (codefile,
		     "p += l; len -= l; ret += l;\n");
	}
	break;
    }
    case TInteger:
	if(t->members) {
	    char *s;
	    asprintf(&s, "(int*)%s", name);
	    if(s == NULL)
		errx (1, "out of memory");
	    decode_primitive ("integer", s, forwstr);
	    free(s);
	} else if (t->range == NULL) {
	    decode_primitive ("heim_integer", name, forwstr);
	} else if (t->range->min == INT_MIN && t->range->max == INT_MAX) {
	    decode_primitive ("integer", name, forwstr);
	} else if (t->range->min == 0 && t->range->max == UINT_MAX) {
	    decode_primitive ("unsigned", name, forwstr);
	} else if (t->range->min == 0 && t->range->max == INT_MAX) {
	    decode_primitive ("unsigned", name, forwstr);
	} else
	    errx(1, "%s: unsupported range %d -> %d", 
		 name, t->range->min, t->range->max);
	break;
    case TBoolean:
      decode_primitive ("boolean", name, forwstr);
      break;
    case TEnumerated:
	decode_primitive ("enumerated", name, forwstr);
	break;
    case TOctetString:
	decode_primitive ("octet_string", name, forwstr);
	break;
    case TBitString: {
	Member *m;
	int pos = 0;

	if (TAILQ_EMPTY(t->members)) {
	    decode_primitive ("bit_string", name, forwstr);
	    break;
	}
	fprintf(codefile,
		"if (len < 1) return ASN1_OVERRUN;\n"
		"p++; len--; ret++;\n");
	fprintf(codefile,
		"do {\n"
		"if (len < 1) break;\n");
	TAILQ_FOREACH(m, t->members, members) {
	    while (m->val / 8 > pos / 8) {
		fprintf (codefile,
			 "p++; len--; ret++;\n"
			 "if (len < 1) break;\n");
		pos += 8;
	    }
	    fprintf (codefile,
		     "(%s)->%s = (*p >> %d) & 1;\n",
		     name, m->gen_name, 7 - m->val % 8);
	}
	fprintf(codefile,
		"} while(0);\n");
	fprintf (codefile,
		 "p += len; ret += len;\n");
	break;
    }
    case TSequence: {
	Member *m;

	if (t->members == NULL)
	    break;

	TAILQ_FOREACH(m, t->members, members) {
	    char *s;

	    asprintf (&s, "%s(%s)->%s", m->optional ? "" : "&", name, m->gen_name);
	    decode_type (s, m->type, m->optional, forwstr);
	    free (s);
	}

	break;
    }
    case TSet: {
	Member *m;
	unsigned int memno;

	if(t->members == NULL)
	    break;

	fprintf(codefile, "{\n");
	fprintf(codefile, "unsigned int members = 0;\n");
	fprintf(codefile, "while(len > 0) {\n");
	fprintf(codefile, 
		"Der_class class;\n"
		"Der_type type;\n"
		"int tag;\n"
		"e = der_get_tag (p, len, &class, &type, &tag, NULL);\n"
		"if(e) %s;\n", forwstr);
	fprintf(codefile, "switch (MAKE_TAG(class, type, tag)) {\n");
	memno = 0;
	TAILQ_FOREACH(m, t->members, members) {
	    char *s;

	    assert(m->type->type == TTag);

	    fprintf(codefile, "case MAKE_TAG(%s, %s, %s):\n",
		    classname(m->type->tag.tagclass),
		    is_primitive_type(m->type->subtype->type) ? "PRIM" : "CONS",
		    valuename(m->type->tag.tagclass, m->type->tag.tagvalue));

	    asprintf (&s, "%s(%s)->%s", m->optional ? "" : "&", name, m->gen_name);
	    if(m->optional)
		fprintf(codefile, 
			"%s = calloc(1, sizeof(*%s));\n"
			"if (%s == NULL) { e = ENOMEM; %s; }\n",
			s, s, s, forwstr);
	    decode_type (s, m->type, 0, forwstr);
	    free (s);

	    fprintf(codefile, "members |= (1 << %d);\n", memno);
	    memno++;
	    fprintf(codefile, "break;\n");
	}
	fprintf(codefile, 
		"default:\n"
		"return ASN1_MISPLACED_FIELD;\n"
		"break;\n");
	fprintf(codefile, "}\n");
	fprintf(codefile, "}\n");
	memno = 0;
	TAILQ_FOREACH(m, t->members, members) {
	    char *s;

	    asprintf (&s, "%s->%s", name, m->gen_name);
	    fprintf(codefile, "if((members & (1 << %d)) == 0)\n", memno);
	    if(m->optional)
		fprintf(codefile, "%s = NULL;\n", s);
	    else if(m->defval)
		gen_assign_defval(s, m->defval);
	    else
		fprintf(codefile, "return ASN1_MISSING_FIELD;\n");
	    free(s);
	    memno++;
	}
	fprintf(codefile, "}\n");
	break;
    }
    case TSetOf:
    case TSequenceOf: {
	char *n;

	fprintf (codefile,
		 "{\n"
		 "size_t origlen = len;\n"
		 "size_t oldret = ret;\n"
		 "ret = 0;\n"
		 "(%s)->len = 0;\n"
		 "(%s)->val = NULL;\n"
		 "while(ret < origlen) {\n"
		 "(%s)->len++;\n"
		 "(%s)->val = realloc((%s)->val, sizeof(*((%s)->val)) * (%s)->len);\n",
		 name, name, name, name, name, name, name);
	asprintf (&n, "&(%s)->val[(%s)->len-1]", name, name);
	decode_type (n, t->subtype, 0, forwstr);
	fprintf (codefile, 
		 "len = origlen - ret;\n"
		 "}\n"
		 "ret += oldret;\n"
		 "}\n");
	free (n);
	break;
    }
    case TGeneralizedTime:
	decode_primitive ("generalized_time", name, forwstr);
	break;
    case TGeneralString:
	decode_primitive ("general_string", name, forwstr);
	break;
    case TTag:{
	fprintf(codefile, 
		"{\n"
		"int reallen;\n");
	fprintf(codefile, "e = der_match_tag_and_length(p, len, %s, %s, %s, "
		"&reallen, &l);\n",
		classname(t->tag.tagclass),
		is_primitive_type(t->subtype->type) ? "PRIM" : "CONS",
		valuename(t->tag.tagclass, t->tag.tagvalue));
	if(optional) {
	    fprintf(codefile, "if(e)\n"
		    "%s = NULL;\n", name);
	    fprintf(codefile, "else {\n");
	    if(dce_fix)
		fprintf (codefile,
			 "int dce_fix;\n");
	    fprintf (codefile,
		     "size_t oldlen;\n"
		     "%s = calloc(1, sizeof(*%s));\n"
		     "if (%s == NULL) { e = ENOMEM; %s; }\n",
		     name, name, name, forwstr);
	} else {
	    fprintf(codefile, "if(e) %s;\n", forwstr);
	    fprintf(codefile, "{\n");
	    if(dce_fix)
		fprintf(codefile, 
			"int dce_fix;\n");
	    fprintf(codefile, 
		    "size_t oldlen;\n");
	}
	fprintf (codefile,
		 "p += l; len -= l; ret += l;\n"
		 "oldlen = len;\n");
	if(dce_fix)
	    fprintf (codefile,
		     "if((dce_fix = _heim_fix_dce(reallen, &len)) < 0)\n"
		     "{ e = ASN1_BAD_FORMAT; %s; }\n",
		     forwstr);
	else
	    fprintf(codefile, 
		    "if (reallen > len) { e = ASN1_OVERRUN; %s; }\n"
		    "len = reallen;\n", forwstr);
	decode_type (name, t->subtype, 0, forwstr);
	if(dce_fix)
	    fprintf(codefile,
		    "if(dce_fix){\n"
		    "e = der_match_tag_and_length (p, len, "
		    "(Der_class)0, (Der_type)0, 0, &reallen, &l);\n"
		    "if(e) %s;\np += l; len -= l; ret += l;\n"
		    "} else \n", forwstr);
	fprintf(codefile, 
		"len = oldlen - reallen;\n"
		"}\n"
		"}\n");
	break;
    }
    case TChoice: {
	Member *m;

	if (t->members == NULL)
	    break;

	fprintf(codefile, "{\nint choice_match = 0;\n");
	TAILQ_FOREACH(m, t->members, members) {
	    char *s;

	    asprintf (&s, "%s(%s)->u.%s", m->optional ? "" : "&", name, m->gen_name);
	    fprintf(codefile, "if(choice_match == 0) {\n");
	    fprintf(codefile, "do {\n");
	    decode_type (s, m->type, m->optional, "break");
	    fprintf(codefile, "(%s)->element = %s;\n", name, m->label);
	    fprintf(codefile, "choice_match = 1;\n");
	    fprintf(codefile, "} while (0);\n");
	    fprintf(codefile, "}\n");
	    free (s);
	}
	fprintf(codefile, "if(choice_match == 0) {\n"
		"e = ASN1_PARSE_ERROR;\n"
		"%s;\n"
		"}\n",
		forwstr);
	fprintf(codefile, "}\n");
	break;
    }
    case TUTCTime:
	decode_primitive ("utctime", name, forwstr);
	break;
    case TUTF8String:
	decode_primitive ("utf8string", name, forwstr);
	break;
    case TPrintableString:
	decode_primitive ("printable_string", name, forwstr);
	break;
    case TIA5String:
	decode_primitive ("ia5_string", name, forwstr);
	break;
    case TNull:
	fprintf (codefile, "/* NULL */\n");
	break;
    case TOID:
	decode_primitive ("oid", name, forwstr);
	break;
    default :
	abort ();
    }
    return 0;
}

void
generate_type_decode (const Symbol *s)
{
  int preserve = preserve_type(s->name) ? TRUE : FALSE;

  fprintf (headerfile,
	   "int    "
	   "decode_%s(const unsigned char *, size_t, %s *, size_t *);\n",
	   s->gen_name, s->gen_name);

  fprintf (codefile, "int\n"
	   "decode_%s(const unsigned char *p,"
	   " size_t len, %s *data, size_t *size)\n"
	   "{\n",
	   s->gen_name, s->gen_name);

  switch (s->type->type) {
  case TInteger:
  case TBoolean:
  case TOctetString:
  case TOID:
  case TGeneralizedTime:
  case TGeneralString:
  case TUTF8String:
  case TPrintableString:
  case TIA5String:
  case TUTCTime:
  case TNull:
  case TEnumerated:
  case TBitString:
  case TSequence:
  case TSequenceOf:
  case TSet:
  case TSetOf:
  case TTag:
  case TType:
  case TChoice:
    fprintf (codefile,
	     "size_t ret = 0, reallen;\n"
	     "size_t l;\n"
	     "int e;\n");
    if (preserve)
	fprintf (codefile, "const unsigned char *begin = p;\n");

    fprintf (codefile, "\n");
    fprintf (codefile, "memset(data, 0, sizeof(*data));\n");
    fprintf (codefile, "reallen = 0;\n"); /* hack to avoid `unused variable' */

    decode_type ("data", s->type, 0, "goto fail");
    if (preserve)
	fprintf (codefile,
		 "data->_save.data = calloc(1, ret);\n"
		 "if (data->_save.data == NULL) { e = ENOMEM; goto fail; }\n"
		 "data->_save.length = ret;\n"
		 "memcpy(data->_save.data, begin, ret);\n");
    fprintf (codefile, 
	     "if(size) *size = ret;\n"
	     "return 0;\n");
    fprintf (codefile,
	     "fail:\n"
	     "free_%s(data);\n"
	     "return e;\n",
	     s->gen_name);
    break;
  default:
    abort ();
  }
  fprintf (codefile, "}\n\n");
}
